package SIPLock;

use warnings;
use Sys::Hostname;
use English;
use strict;
use IPTables::ChainMgr;
use DBD::SQLite;
use DBI;
use Sys::Syslog;
use Class::InsideOut qw( public private register id );
use Data::Dumper;
use LWP::Simple;
use Net::IP::Match::Regexp qw( create_iprange_regexp );
use Geo::IP::PurePerl;
use Digest::MD5 qw(md5_hex);

use constant DB_NAME		=>	'/var/lib/asterisk/SIPLock.db';
use constant LICENSE_URL	=>	'https://ps.fonality.com/Market/validate_license.cgi?product_id=4';

my %config;
my %iptable;
my %dbinst;

eval('public config	 => %config;');
eval('public iptable => %iptable;');
eval('public dbinst	 => %dbinst;');

openlog('SIPLock', "pid", "local0");

sub new {
  my $package = shift;
  my $new_object = bless({}, $package);
  register($new_object);
  die "No database!" if (! -e DB_NAME);
  $dbinst{id $new_object} = DBI->connect("dbi:SQLite:dbname=".DB_NAME, q{}, q{}, { RaiseError => 1, AutoCommit => 1, PrintError => 0 }) || die $!;
  
  my %opts = (
      'iptables' => '/sbin/iptables', # can specify 'ip6tables' hash key instead
      'iptout'   => '/tmp/iptables.out',
      'ipterr'   => '/tmp/iptables.err',
      'debug'    => 0,
      'verbose'  => 0,

      ### advanced options
      'ipt_alarm' => 5,  ### max seconds to wait for iptables execution.
      'ipt_exec_style' => 'waitpid',  ### can be 'waitpid',
                                      ### 'system', or 'popen'.
      'ipt_exec_sleep' => 1, ### add in time delay between execution of
                             ### iptables commands (default is 0).
  );
  my $rv = 0;
  my $out_ar = [];
  my $errs_ar = [];
  
  $iptable{id $new_object} = new IPTables::ChainMgr(%opts) or warn $!;
  
  # Clear all iptables rules involving udp, port 5060, drop
  ($rv, $out_ar, $errs_ar) = $iptable{id $new_object}->chain_exists('filter', 'SIPLock');
  if ($rv) {
    $iptable{id $new_object}->flush_chain('filter', 'SIPLock');
	$iptable{id $new_object}->delete_chain('filter', 'INPUT', 'SIPLock');
  }
  $iptable{id $new_object}->create_chain('filter', 'SIPLock');
  $iptable{id $new_object}->add_jump_rule('filter', 'INPUT', 1, 'SIPLock');

  my $hn = hostname;
  if ($hn=~m/^(pbxtra|trixbox)(\d+)$/) {
    $config{id $new_object}{'serverID'} = $2;
  } else {
    $config{id $new_object}{'serverID'} = "";
  }
  $config{id $new_object}{'currentBlocks'}{0}=0;
  $config{id $new_object}{'sipPort'} = getservbyport("5060","") || "5060";
  getConfig($new_object);
  return $new_object;
}

sub debug {
  my $self = shift;
  if (@_) {
    $self->{'debug'} = shift;
  }
  return $self->{'debug'};
}

sub ResetFW {
  my ($self) = @_;
  Log($self,"ResetFW()");
  
  # Find the saved entries and set them up.
  my $all = $dbinst{id $self}->prepare("SELECT ip FROM blocked_hosts");
  $all->execute();
  while (my ($addr) = $all->fetchrow_array()) {
	AddEntrySilent($self, $addr);
    $config{id $self}{'currentBlocks'}{$addr}=1;
    Log($self, "ResetFW(): Added ". $addr);
  }
  ResetWhiteBlackLists($self);
}

sub FindRule {
  my ($self, $net) = @_;
  $net=~s/\/32//;
  
  my ($rulenum, $chain_rules) = $iptable{id $self}->find_ip_rule($net,'0.0.0.0/0','filter', 'SIPLock', 'REJECT', {});

  return($rulenum);
}

sub ResetWhiteBlackLists {
  my ($self) = @_;
  Log($self,"ResetWhiteBlackLists()");
  my $doCommitIP=0;
  my $all = $dbinst{id $self}->prepare("SELECT net, mask, type, deleteme FROM exception_list order by type");
  $all->execute();
  while (my ($addr, $mask, $type, $deleteme) = $all->fetchrow_array()) {
    my $oldaddr=$addr;
    $addr.="/".$mask;
    if ($deleteme == 0) {
      # New or existing whitelist or blacklist entry.

      if ( ($type == 1) && (!defined($config{id $self}{'whitelist'}{$addr})) ) {
        $config{id $self}{'whitelist'}{$addr}=1;
        $doCommitIP++;
        Log($self, "ResetFW(): Added whitelist entry ". $addr);
        $dbinst{id $self}->do("UPDATE exception_list SET complete=1 WHERE net='$oldaddr' AND mask=$mask AND deleteme=0");
      } elsif ( ($type == 2) && (! FindRule($self, $addr)) ) {
		AddEntrySilent($self, $addr);
        Log($self, "ResetFW(): Added blacklist entry ". $addr);
        $dbinst{id $self}->do("UPDATE exception_list SET complete=1 WHERE net='$oldaddr' AND mask=$mask AND deleteme=0");
      }
    } else {
      # Delete a whitelist or blacklist entry.
      $dbinst{id $self}->do("DELETE FROM exception_list WHERE net='$oldaddr' AND mask=$mask AND deleteme=1");

      if ($type == 1) {
        delete ($config{id $self}{'whitelist'}{$addr});
        $doCommitIP++;
        Log($self, "ResetFW(): Removed whitelist entry ". $addr);
      } else {
        DelEntrySilent($self, $addr);
        Log($self, "ResetFW(): Removed blacklist entry ". $addr);
      }

    }
  }

  # Manually deleted automatic rules.
  $all = $dbinst{id $self}->prepare("SELECT ip FROM blocked_hosts WHERE deleteme=1");
  $all->execute();
  while (my ($addr) = $all->fetchrow_array()) {
    $dbinst{id $self}->do("DELETE FROM blocked_hosts WHERE ip='$addr' AND deleteme=1");
	DelEntrySilent($self, $addr);
    delete($config{id $self}{'currentBlocks'}{$addr});
    Log($self, "ResetFW(): User requested removal of autoblock entry ". $addr);
    shareTransaction($self, $addr, undef, "U");
    Transaction($self, $addr, "2");
  } 

  if ($doCommitIP) {
    my @ips = keys %{$config{id $self}{'whitelist'}};
	if (@ips) {
      $config{id $self}{'WhiteListRegex'} = create_iprange_regexp( keys %{$config{id $self}{'whitelist'}});
	} else {
	  $config{id $self}{'WhiteListRegex'} = [];
    }	
  }
}

sub AddEntry {
  my ($self, $addr) = @_;
  DelEntrySilent($self, $addr);
  doAddEntry($self, $addr, 0);
}

sub AddEntrySilent {
  my ($self, $addr) = @_;
  return(doAddEntry($self, $addr, 1));
}

sub doAddEntry {
  my ($self, $addr, $silent) = @_;
  my @ret_val = $iptable{id $self}->append_ip_rule($addr, '0.0.0.0/0', 'filter', 'SIPLock', 'REJECT', 
    {'protocol' => 'udp'});
  
  if ($ret_val[0]) {
    Log($self, "Added $addr") if (! $silent);
  } else {
    Log($self, "Failure to add $addr") if (! $silent);
  }
}

sub DelEntry {
  my ($self, $addr) = @_;
  return(doDelEntry($self, $addr, 0));
}

sub DelEntrySilent {
  my ($self, $addr) = @_;
  return(doDelEntry($self, $addr, 1));
}

sub doDelEntry {
  my ($self, $addr, $silent) = @_;
  my @ret_val = $iptable{id $self}->delete_ip_rule($addr, '0.0.0.0/0', 'filter', 'SIPLock', 'REJECT', 
    {'protocol' => 'udp'});
	
  if ($ret_val[0]) {
    Log($self, "Removed $addr") if (! $silent);
  } else {
    $addr =~ /(\/\d+)/;
	# Lets try without the sub mask
	if ($`) {
	  doDelEntry($self, $`, $silent);
	} else {
      Log($self, "Failure to remove $addr") if (! $silent);
	}  
  }
}

sub getConfig {
  my ($self) = @_;
  my $all = $dbinst{id $self}->prepare("SELECT * FROM settings");
  $all->execute();
  while (my @row = $all->fetchrow_array()) {
    $config{id $self}{$row[0]} = $row[1];
  } 
}

sub sqlAddEntry {
  my ($self, $addr) = @_;
  no warnings 'all';
  eval { $dbinst{id $self}->do("INSERT INTO blocked_hosts (ip, country) VALUES ('$addr', '".getCountry($self, $addr)."')"); };
}

sub sqlDelEntry {
  my ($self, $addr) = @_;
  $dbinst{id $self}->do("DELETE FROM blocked_hosts WHERE ip='$addr'");
}

sub Log {
  my ($self, $msg) = @_;
  return if (! $self->{'debug'});
  syslog("info|local0", $msg);
}

sub LogDie {
  my ($self, $msg) = @_;
  Log($self, $msg);
  die ($msg);
}

sub Transaction {
  my ($self, $ip, $type) = @_;
  $dbinst{id $self}->do("INSERT INTO transactions (ip, type, country) VALUES ('$ip', $type, '". getCountry($self, $ip)."')");
}

sub shareTransaction {
  my ($self, $ip, $user, $type) = @_;
  if (defined($user)) {
    Log($self, "Shared $type transaction from ". $config{id $self}{'serverID'} ." at $ip with authentication @{$user}");
    get("https://siplock.fonality.com/SIPLock.cgi?sid=".$config{id $self}{'serverID'} ."&ip=".$ip."&type=".$type."&user=@{$user}");
  } else {
    Log($self, "Shared $type transaction from ". $config{id $self}{'serverID'} ." at $ip");
    get("https://siplock.fonality.com/SIPLock.cgi?sid=".$config{id $self}{'serverID'} ."&ip=".$ip."&type=".$type);
  }
  return(1);
}

sub getCountry {
  my ($self, $ip) = @_;

  my $geo = Geo::IP::PurePerl->open("/usr/local/fonality/prosvcs/GeoIP/GeoLiteCity.dat", GEOIP_STANDARD);
  my $location = $geo->get_city_record_as_hash($ip);
  return("UNKNOWN") if (!$location->{'country_name'});

  my $location_string;
  $location_string .= $location->{'city'} . ', ' if $location->{'city'};
  $location_string .= $location->{'region'} . ', ' if($location->{'region'} && $location->{'region'} !~ /^\d+$/);
  $location_string .= $location->{'country_name'};
  return($location_string);
}

sub checkLicense {
  my ($self) = @_;
  return(1) if (get(LICENSE_URL."&server_id=".$config{id $self}{'serverID'}) eq 'Success');
  return(0);
}

1;
__END__
